-- SyncTradesTest.lua
running = true                -- Script operation flag
TRADE_ACC="4100R24"           -- Trading account
CLASS_CODE_FUT="SPBFUT"       -- Futures class
SEC_CODE_FUT_FOR_OPEN="SiU6"  -- Security (futures) code
TRANS_ID=0                    	-- Transaction ID set by a user when sending a transaction from QLUA

-- 
local logfile = 0 				-- A file for logging
-- Auxiliary variables
local teststarted = false		-- The flag of a running test
local lastorder_ticket=""		-- The ticket of the last sent order
local lastorderTPSL=""			-- The ticket of the last order to place TP&SL request	
local lastTPSL_trans_id=""		-- The number of the last transaction to place the TP&SL request	
local last_operation="No"		-- The type of the last transaction
local trades_limit=10;			-- The number of trades in a series
-- Time counter
local start,finish			  	-- Testing beginning and end time
local topen_start,topen_finish	-- Beginning and confirmation of a buy operation
local tclose_start,tclose_stop	-- Beginning and confirmation of a sell operation
local torder_start
local torder_finish 			-- Beginning and confirmation of a conditional request placing
local tdelete_start
local tdelete_finish 			-- Beginning and confirmation of a conditional request deletion
local timeOpen,timeClose 		-- Total time spent for buy/sell operations, for finding the average value
local timeSet, timeDelete		-- Total time spent for place/cancel operations, for finding the average value
local tradecounter              -- Counter of performed trades
local transIDstart				-- The starting transaction ID is set based on the minute and second

-- Automation state
local ready="READY"    			-- waiting for a command
local startSeries="seriesStarted"-- a series of trading operations has started
local buyDone="buyDone"  		-- buy operation performed
local SetOrder="SetOrder" 		-- TP&SL order placed
local sellDone="sellDone"		-- sell operation performed
local deleteDone="deleteDone"	-- TP&SL order canceled
local seriesDone="seriesDone"	-- a series of trade operation completed
local noChanges="noChanges"		-- state unchanged
local state=ready
-- Operation types
local BUY="BUY"
local SELL="SELL"
local SETORDER="SETORDER"
local DELETEORDER="DELETEORDER"

-- Auxiliary variables in order to avoid multiple calls of OnTrade() and OnOrder()
local ontrade_event=""			-- line trans_id+order_num
local onorder_event=""			-- line trans_id+order_num

-- This is the main function if the script, runs automatically; the script runs until the function completes 
function main()
	-- Tries to open a file in the "read/write" mode
	logfile = io.open(getScriptPath().."\\SyncTradesTest_log.txt","a+");
	-- If the file does not exist
	if logfile == nil then 
		-- Creates a file in the "write" mode
		logfile = io.open(getScriptPath().."\\SyncTradesTest_log.txt","w"); 
		-- Closes the file
		logfile:close();
		-- Opens an existing file and the "read/write" mode
		logfile = io.open(getScriptPath().."\\SyncTradesTest_log.txt","r+");
	end;
	-- The flag is only switched when the script is stopped by a user
    while running do
		--message(os.date().. "running SyncTradesTest.lua") -- shows date ant time once in 15 seconds
        sleep(15000)
    end
	logfile:flush();
	logfile:close();
end

-- Called with each update of the order book
function OnQuote(class, sec ) 
	-- Starting a series of trades
	if (teststarted==false) then
		teststarted=true;
		message("SyncTradesTest.lua starting");
		Print("SyncTradesTest.lua starting...");
		StartSeries(trades_limit);   -- Starting counters of transactions and time
		start=os.clock();			  -- Start time
		Buy()
	end
end

-- Buying the instrument
function Buy()
	transIDstart=transIDstart+1
    TRANS_ID = transIDstart
	Print((tradecounter+1)..". Buy".." TRANS_ID="..TRANS_ID)
   -- Fills a structure to send a transaction to buy 1 lot 
    local Transaction={
      ["TRANS_ID"]   = tostring(TRANS_ID),
      ["ACTION"]     = "NEW_ORDER",
      ["CLASSCODE"]  = CLASS_CODE_FUT,
      ["SECCODE"]    = SEC_CODE_FUT_FOR_OPEN,
      ["OPERATION"]  = "B", -- BUY
      ["TYPE"]       = "M", -- MARKET
      ["QUANTITY"]   = "1", -- amount
      ["ACCOUNT"]    = TRADE_ACC,
      ["PRICE"]      = tostring(getParamEx(CLASS_CODE_FUT, SEC_CODE_FUT_FOR_OPEN, "offer").param_value + 30*getParamEx(CLASS_CODE_FUT, SEC_CODE_FUT_FOR_OPEN, "SEC_PRICE_STEP").param_value), -- at a price 30 minimal steps higher
      ["COMMENT"]    = string.format("Trade #d Buy %s at %s",(tradecounter+1),SEC_CODE_FUT_FOR_OPEN,os.date("%Y.%m.%d %H:%M:%S"))
    }
    -- At what price to buy
        local price=getParamEx(CLASS_CODE_FUT, SEC_CODE_FUT_FOR_OPEN, "offer").param_value
    --message("Offer Price="..price)
    -- Time of sending a trade transaction to a server
	topen_start=os.clock();  -- The time of buying
	last_operation=BUY
    local Result = sendTransaction(Transaction);
    -- Checking correctness
	if Result ~= "" then
		message("Buying failed!\nERROR: "..tostring(Result));
    end;    
end

-- Selling the instrument
function Sell()
	transIDstart=transIDstart+1
    TRANS_ID = transIDstart
	Print((tradecounter+1)..". Sell".." TRANS_ID="..TRANS_ID)
   -- Fills a structure to send a transaction to buy 1 lot 
    local Transaction={
      ["TRANS_ID"]   = tostring(TRANS_ID),
      ["ACTION"]     = "NEW_ORDER",
      ["CLASSCODE"]  = CLASS_CODE_FUT,
      ["SECCODE"]    = SEC_CODE_FUT_FOR_OPEN,
      ["OPERATION"]  = "S", -- покупка (BUY)
      ["TYPE"]       = "M", -- MARKET
      ["QUANTITY"]   = "1", -- amount
      ["ACCOUNT"]    = TRADE_ACC,
      ["PRICE"]      = tostring(getParamEx(CLASS_CODE_FUT, SEC_CODE_FUT_FOR_OPEN, "bid").param_value - 30*getParamEx(CLASS_CODE_FUT, SEC_CODE_FUT_FOR_OPEN, "SEC_PRICE_STEP").param_value), -- at a price 30 minimal steps higher
      ["COMMENT"]    = string.format("Trade #d Sell %s at %s",(tradecounter+1),SEC_CODE_FUT_FOR_OPEN,os.date("%Y.%m.%d %H:%M:%S"))
    }
    -- At what price to sell
	local price=getParamEx(CLASS_CODE_FUT, SEC_CODE_FUT_FOR_OPEN, "bid").param_value
    --message("Bid Price="..price)
    -- Time of sending a trade transaction to a server
	tclose_start=os.clock();  -- The time of selling
	last_operation=SELL	
    local Result = sendTransaction(Transaction);
    -- Checking correctness
	if Result ~= "" then
		message("Selling failed!\nERROR: "..tostring(Result));
    end;    
end

--
function sendOrderTP_SL()
	transIDstart=transIDstart+1
    TRANS_ID = transIDstart
	Print((tradecounter+1)..". Send TP&SL".." TRANS_ID="..TRANS_ID)
	
    -- At what price to sell
	local price=getParamEx(CLASS_CODE_FUT, SEC_CODE_FUT_FOR_OPEN, "bid").param_value
    --message("Bid Price="..price)
	priceStop = tostring(math.floor(getParamEx(CLASS_CODE_FUT, SEC_CODE_FUT_FOR_OPEN, 'PRICEMAX').param_value)); -- After the triggering of a Stop, the price is maximum possible in order to avoid slippage
	-- Fills a structure to send a transaction to buy 1 lot 
    local Transaction={
		["TRANS_ID"]   = tostring(TRANS_ID),
		["ACTION"]     = "NEW_ORDER",
		["CLASSCODE"]  = CLASS_CODE_FUT,
		["SECCODE"]    = SEC_CODE_FUT_FOR_OPEN,
		["ACCOUNT"]    = TRADE_ACC,		
		["QUANTITY"] = "1",
		["STOP_ORDER_KIND"] = "TAKE_PROFIT_AND_STOP_LIMIT_ORDER",
		["OPERATION"] = "S", -- for short positions "B" - BUY, for long - "S" (SELL))
		["STOPPRICE"] = tostring(price+200), -- Take Profit activation price
		["OFFSET"] = "50",
		["OFFSET_UNITS"] = "PRICE_UNITS",
		["MARKET_TAKE_PROFIT"] = "YES",
		["PRICE"] = priceStop,
		["STOPPRICE2"] = tostring(price-200), --stop price
		["IS_ACTIVE_IN_TIME"] = "YES",
		["ACTIVE_FROM_TIME"] = "100000",   -- Hours:Minutes:Seconds
		["ACTIVE_TO_TIME"] = "234545",     -- Hours:Minutes:Seconds
		["MARKET_STOP_LIMIT"] = "YES",
		["COMMENT"]    = string.format("Trade #d Set TP&SL for %s at %s",(tradecounter+1),SEC_CODE_FUT_FOR_OPEN,os.date("%Y.%m.%d %H:%M:%S"))
    }
    -- Time of sending a trade transaction to a server
	torder_start=os.clock();  -- time of sending a conditional order
	last_operation=SETORDER	
    local Result = sendTransaction(Transaction);
    -- Checking correctness
	if Result ~= "" then
		message("Placing TP&SL order failed!\nERROR: "..tostring(Result));
    end;
end;

-- Deleting a pending order
function DeleteOrder()
	--transIDstart=transIDstart+1
	TRANS_ID = lastTPSL_trans_id -- transIDstart
	Print((tradecounter+1)..". Delete TP&SL #"..lastorderTPSL.." TRANS_ID="..TRANS_ID)
	-- Fills a structure to cancel a transaction order 
    local Transaction={
		["TRANS_ID"]   = tostring(TRANS_ID),
		["ACTION"]     = "KILL_ORDER",
		["CLASSCODE"]  = CLASS_CODE_FUT,
		["SECCODE"]    = SEC_CODE_FUT_FOR_OPEN,
		["ACCOUNT"]    = TRADE_ACC,		
		["ORDER_KEY"] = tostring(lastorderTPSL),
		["COMMENT"]    = string.format("Trade #d Delete order %d at %s",(tradecounter+1),lastorderTPSL,os.date("%Y.%m.%d %H:%M:%S"))
    }
    -- Time of sending a trade transaction to a server
	tdelete_start=os.clock();  -- Time of sending an order to delete a conditional request 
	last_operation=DELETEORDER	
    local Result = sendTransaction(Transaction);
    -- Checking correctness
	if Result ~= "" then
		message("Deleting a TP&SL order #"..lastorderTPSL.." failed!\nERROR: "..tostring(Result));
    end;
end;

-- Initiate counters
function StartSeries(total_trades)
	trades_limit=total_trades
	timeOpen=0
	timeClose=0
	timeSet=0
	timeDelete=0
	tradecounter=0; 	
	minutes=os.date("%M")
	seconds=os.date("%S")
	transIDstart=minutes*10000+seconds*100
end

-- Called when a new stop order is received of the parameters of an existing stop order are modified
function OnTrade(order)
    --Print(tradecounter..". OnTrade #"..order.order_num.." Flags="..order.flags..order.system_ref) 
	-- Only events of the last transaction are handled
	if(order.trans_id~=TRANS_ID) then	
		return -- other events (e.g. late) should be skipped, since they break the logic
	end
	-- The event can arrive 3 times, handle only the first call
	curr_ontrade_event=order.trans_id..order.order_num
	if (curr_ontrade_event~=ontrade_event) then
		ontrade_event=curr_ontrade_event
		-- Check the current automation state
		state=CheckState(order, nil, nil);
		--Print("state="..tostring(state));
		-- Process the state
		ProcessState(state)
	end
end

-- Called when receiving a response to a user's transaction sent using Trans2quik.dll, QPILE, QLua 
function OnTransReply(trans_reply)
	-- Only events of the last transaction are handled
	if(trans_reply.trans_id~=TRANS_ID) then	
		return -- other events (e.g. late) should be skipped, since they break the logic
	end
    state=CheckState(nil, trans_reply, nil)
	-- Process the state
	--ProcessState(state)	
end

-- Called by the QUIK terminal when a new order is received of the parameters of an existing order are modified
function OnOrder(order)
	-- Only events of the last transaction are handled
	if(order.trans_id~=TRANS_ID) then	
		return -- other events (e.g. late) should be skipped, since they break the logic
	end
	-- The event can arrive 3 times, handle only the first call
	curr_onorder_event=order.trans_id..order.order_num
	if (curr_onorder_event~=onorder_event) then
	    onorder_event=curr_onorder_event
		state=CheckState(nil, nil, order)
		-- Process the state
		ProcessState(state)		
	end
end


-- Perform actions in accordance with the state
function ProcessState(currstate)
	-- Searching through states
	if(currstate==startSeries) then -- Test started - first buying performed	
		start=os.clock();			-- Start time
		Buy();                  	-- Buying
	end;
	if(currstate==buyDone)  then  	-- Long position has opened
         Sell(); 			   		-- Closing the position // sending the order "Take Profit and Stop Limit"
	end;
	if(currstate==deleteDone) then	-- Conditional order deleted, a new position can be opened
         Buy();         			-- Buying by market
	end;
	if(currstate==setDone) then     -- SL and TP placed
         Sell();    	   			-- Closing the position
	end;
	if(currstate==sellDone) then	-- position closed, we should delete the "Take Profit and Stop Limit" order
         Buy();						-- Buying by market
	end;
	if(currstate==seriesDone) then  -- The last one of TradeSeries positions is closed 
		finish=os.clock();  	-- Completion time 
		delta=(finish-start)*1000;
		Print(string.format("Total time=%.1f ms",delta));
		Print(string.format("Average time: Open=%.1f ms, Close %.1f ms (%d series)",
                     timeOpen*1000/tradecounter,timeClose*1000/tradecounter,tradecounter));
		final_msg=string.format("Total series=%d, total time=%.1f ms",tradecounter,delta)
		final_msg=final_msg.."\n"..string.format("Average time: open=%.1f ms, close %.1f ms (%d series)",
                     timeOpen*1000/tradecounter,timeClose*1000/tradecounter,tradecounter)
		message(final_msg);
		-- Setting the flag to 'Off'
		--running = false
		--Print("finishing...");		
	end
	--info=getOnTradeStructure(order,tradecounter);
	--logfile:write(info);
end

-- Checking the state
function CheckState(orderfromOnTrade, transfromOnTransreply, orderfromOnOrder)
    temp_state=noChanges
    -- Processing the call from OnTransReply	
	if (transfromOnTransreply~=nil) then	
		TransID=transfromOnTransreply.trans_id	-- transaction id
		Flags=transfromOnTransreply.flags
		lastorder_ticket=transfromOnTransreply.order_num
		-- Getting the transaction status
		status=getSatus(transfromOnTransreply.status)
		-- Write to log
		Print(string.format("OnTransReply: TransID=%d  %s Flags=%d",TransID,status,Flags))
	end

    -- Processing the call from OnTrade
	if (orderfromOnTrade~=nil) then
		-- Order filled, a trade appeared in the trading system
		stringdeal=tostring(orderfromOnTrade.trade_num)
		order=orderfromOnTrade.order_num		-- order ticket
		TransID=orderfromOnTrade.trans_id		-- transaction id
		Flags=orderfromOnTrade.flags
		-- Write to log
		Print(string.format("OnTrade: TransID=%d Order_num=%s  Flags=%d",TransID,order,Flags))
		--Print(string.format("%d. TransID=%d order #%s %s %s",(tradecounter+1),TransID, order,last_operation,stringdeal))
		return (noChanges) -- 
	end
    
    -- Processing the call from OnOrder		
	if (orderfromOnOrder~=nil) then	
		order=orderfromOnOrder.order_num		-- order ticket
		TransID=orderfromOnOrder.trans_id		-- transaction id
		Flags=orderfromOnOrder.flags
		-- Write to log
		Print(string.format("OnOrder: TransID=%d Order_num=%s  Flags=%d",TransID,order,Flags))
		
		-- the last transaction was buying		
		if ((orderfromOnOrder.order_num==lastorder_ticket)and(last_operation==BUY)and(orderfromOnOrder.trade_num~=0)) then 
			topen_finish=os.clock()
			timeOpen=timeOpen+(topen_finish-topen_start)
			lastorder_ticket=0
			last_operation="No"
			Print(string.format("%d. Open time=%.1f ms",(tradecounter+1),(topen_finish-topen_start)*1000))
			-- buying performed 
            Print("Returning from CheckState state=buyDone")
			return (buyDone)
		end			
		
		-- the last transaction was selling
		if ((orderfromOnOrder.order_num==lastorder_ticket)and(last_operation==SELL)and(orderfromOnOrder.trade_num~=0)) then 
			tclose_finish=os.clock()
			timeClose=timeClose+(tclose_finish-tclose_start)		
			lastorder_ticket=0
			last_operation="No"
			Print(string.format("%d. Close time=%.1f ms",(tradecounter+1),(tclose_finish-tclose_start)*1000))
			-- selling performed 
			tradecounter=tradecounter+1
			- A series of trades completed
			if(tradecounter>=trades_limit) then
				return(seriesDone) -- Completing the test
			end			
			return (sellDone)
		end	
		-- the last transaction was "TAKE_PROFIT_AND_STOP_LIMIT_ORDER"
		if ((orderfromOnOrder.order_num==lastorder_ticket)and(last_operation==SETORDER)) then 
			torder_finish=os.clock()
			timeSet=timeSet+(torder_finish-torder_start)
			lastorderTPSL=orderfromOnOrder.order_num
			lastTPSL_trans_id=orderfromOnOrder.trans_id
			lastorder_ticket=0
			last_operation="No"
			Print(string.format("%d. SetOrder time=%.1f ms",(tradecounter+1),(torder_finish-torder_start)*1000))
			-- "TAKE_PROFIT_AND_STOP_LIMIT_ORDER" order placed
			return (setDone) 
		end

		-- the last transaction was "KILL_ORDER"
		if ((orderfromOnOrder.order_num==lastorder_ticket)and(last_operation==DELETEORDER)) then 
			tdelete_finish=os.clock()
			timeDelete=timeDelete+(tdelete_finish-tdelete_start)
			lastorder_ticket=0
			last_operation="No"
			Print(string.format("%d. DeleteOrder time=%.1f ms",(tradecounter+1),(tdelete_finish-tdelete_start)*1000))
			tradecounter=tradecounter+1
			- A series of trades completed
			if(tradecounter>=trades_limit) then
				return(seriesDone) -- Completing the test
			end
			-- "TAKE_PROFIT_AND_STOP_LIMIT_ORDER" canceled
			return (deleteDone) 
		end					
	end		
	Print("Returning from CheckState state="..temp_state)
	return (temp_state)
end


-- The event of script stop by user
function OnStop(s)
	-- Setting the flag to 'Off'
	running = false
	Print("SyncTradesTest.lua finishing...");
	return 3000 -- Setting timeout of 3 seconds
end

function Print(line)
	if (logfile~=nil) then
	    timestamp=os.date("%Y.%m.%d %H:%M:%S ");
		logfile:write(timestamp.."\t"..line.."\n");
	end
end

function getSatus(statuscode)
	status_string="Unknown status "..statuscode
	if (statuscode==0) then
		status_string="Transaction sent to a server"
	end
	if (statuscode==1) then
		status_string="Transaction from a client received by the QUIK server"
	end
	if (statuscode==2) then
		status_string="Error occurred while transferring a transaction to the trading system"
	end
	if (statuscode==3) then
		status_string="Trsnaction executed"
	end
	if (statuscode==4) then
		status_string="Transaction not executed by the system"
	end
	if (statuscode==5) then
		status_string="The transaction has not passed check on the QUIK server for some criteria"
	end
	if (statuscode==6) then
		status_string="The transaction has not passed limit check on the QUIK server"
	end
	if (statuscode==10) then
		status_string="Trsnaction not supported by the trading system"
	end
	if (statuscode==11) then
		status_string="The transaction's digital signature has not been verified on the QUIK server"
	end
	return status_string
end 

function getOnTradeStructure(trade, k)
structure=""
structure=os.clock()..structure..k..". OnTrade\n"
structure=structure.."   trade_num="..trade.trade_num.."\n"
structure=structure.."   canceled_datetime="..tostring(trade.canceled_datetime).."\n"
structure=structure.."   order_num="..trade.order_num.."\n"
structure=structure.."   brokerref ="..trade.brokerref .."\n"
structure=structure.."   userid="..trade.userid.."\n"
structure=structure.."   firmid="..trade.firmid.."\n"
structure=structure.."   account="..trade.account  .."\n"
structure=structure.."   flags="..trade.flags.."\n"
structure=structure.."   system_ref="..trade.system_ref.."\n"
structure=structure.."   kind ="..trade.kind.."\n"
structure=structure.."   uid ="..trade.uid.."\n"
return structure
end